/**
 * OW2 Util
 * Copyright (C) 2008 Bull S.A.S.
 * Contact: easybeans@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 *
 * --------------------------------------------------------------------------
 * $Id: EnhancedPool.java 5512 2010-05-27 15:14:57Z benoitf $
 * --------------------------------------------------------------------------
 */

package org.ow2.util.pool.impl.enhanced;

import java.util.concurrent.Executor;

import org.ow2.util.log.Log;
import org.ow2.util.log.LogFactory;
import org.ow2.util.pool.api.IPoolConfiguration;
import org.ow2.util.pool.api.IPoolState;
import org.ow2.util.pool.api.PoolException;
import org.ow2.util.pool.impl.enhanced.api.IPool;
import org.ow2.util.pool.impl.enhanced.api.IWaitControl;
import org.ow2.util.pool.impl.enhanced.api.TimeoutPoolException;
import org.ow2.util.pool.impl.enhanced.api.basic.IBasicPool;
import org.ow2.util.pool.impl.enhanced.api.basic.PoolFactoryBroken;
import org.ow2.util.pool.impl.enhanced.api.keepbusy.ShareMethod;
import org.ow2.util.pool.impl.enhanced.api.resizer.IResizablePool;
import org.ow2.util.pool.impl.enhanced.callback.CreateCB;
import org.ow2.util.pool.impl.enhanced.callback.RemoveCB;
import org.ow2.util.pool.impl.enhanced.impl.basic.BasicPool;
import org.ow2.util.pool.impl.enhanced.impl.basic.BasicPoolState;
import org.ow2.util.pool.impl.enhanced.impl.basic.accessmanager.LastUsedAccessManager;
import org.ow2.util.pool.impl.enhanced.impl.keepbusy.KeepBusyPool;
import org.ow2.util.pool.impl.enhanced.impl.limited.LimitedWaiterPool;
import org.ow2.util.pool.impl.enhanced.impl.limited.TooManyWaiterException;
import org.ow2.util.pool.impl.enhanced.impl.listener.resizer.StatsResizerPool;
import org.ow2.util.pool.impl.enhanced.impl.listener.statistics.BasicPoolStats;
import org.ow2.util.pool.impl.enhanced.impl.listener.statistics.BasicPoolStatsFillerPoolListener;
import org.ow2.util.pool.impl.enhanced.impl.util.ExecutorProvider;
import org.ow2.util.pool.impl.enhanced.impl.validator.ValidatorPool;
import org.ow2.util.pool.impl.enhanced.impl.waitcontrol.ExpireWaitControl;
import org.ow2.util.pool.impl.enhanced.impl.waitcontrol.InfiniteWaitControl;
import org.ow2.util.pool.impl.enhanced.impl.waitcontrol.WaitAuthorization;
import org.ow2.util.pool.impl.enhanced.internal.resizer.api.IResizer;
import org.ow2.util.pool.impl.enhanced.internal.resizer.api.shared.ISharedManager;
import org.ow2.util.pool.impl.enhanced.internal.resizer.impl.AsynchronousResizer;
import org.ow2.util.pool.impl.enhanced.internal.resizer.impl.shared.SharedResizer;
import org.ow2.util.pool.impl.enhanced.internal.resizer.impl.stocker.MinMaxExpectedSpareStockerResizer;
import org.ow2.util.pool.impl.enhanced.manager.IPoolManager;
import org.ow2.util.pool.impl.enhanced.manager.optional.IPoolItemRemoveManager;
import org.ow2.util.pool.impl.enhanced.manager.optional.IValidatorPoolManager;

/**
 *
 * @author Gael Lalire
 * @param <E>
 */
public class EnhancedPool<E> {

    private static final Log LOG = LogFactory.getLog(EnhancedPool.class);

    private IBasicPool<E> basicPool;

    private IPool<E> lastPool;

    private LimitedWaiterPool<E> limitedWaiterPool;

    private StatsResizerPool<E, BasicPoolStats> statsResizerPool;

    private KeepBusyPool<E> keepBusyPool;

    private MinMaxExpectedSpareStockerResizer<BasicPoolStats> minMaxExpectedSparePoolResizer;

    private long timeout;

    private int maxWaiter = LimitedWaiterPool.NO_LIMIT_WAITER;

    private boolean started = true;

    private Object mutex = new Object();

    private RemoveCB<E> removeCB;

    private CreateCB<E> createCB;

    /**
     * Volatile flag to avoid use of synchronization when this reference change.
     */
    private volatile WaitAuthorization waitAuthorization = new WaitAuthorization();

    public EnhancedPool(final IPoolManager<E> poolManager) {
        this(poolManager, ResizerType.SIMPLE, null, ExecutorProvider.NEW_THREAD_EXECUTOR);
    }

    public EnhancedPool(final IPoolManager<E> poolManager, final ResizerType resizerType, final ISharedManager sharedManager,
            final Executor createAndRemoveItemExecutor) {
        LimitedWaiterPool<E> limitedWaiterPool;
        KeepBusyPool<E> keepBusyPool;
        ValidatorPool<E> validatorPool;
        IResizablePool<E> pool;
        IPool<E> currentPool;

        createCB = new CreateCB<E>(poolManager);
        if (IPoolItemRemoveManager.class.isInstance(poolManager)) {
            removeCB = new RemoveCB<E>((IPoolItemRemoveManager<E>) poolManager);
        }
        // 1 clue pool
        basicPool = new BasicPool<E>(createCB, 0, new LastUsedAccessManager<E>(), createAndRemoveItemExecutor,
                removeCB, createAndRemoveItemExecutor);
        pool = basicPool;

        minMaxExpectedSparePoolResizer = new MinMaxExpectedSpareStockerResizer<BasicPoolStats>(0, 0,
                IPoolConfiguration.DEFAULT_SPARE, true);

        IResizer<BasicPoolStats> resizer = minMaxExpectedSparePoolResizer;

        switch (resizerType) {
        case ASYNCHRONOUS:
            resizer = new AsynchronousResizer<BasicPoolStats>(minMaxExpectedSparePoolResizer);
            break;
        default:
        case SHARED_ASYNCHRONOUS:
            resizer = new SharedResizer<BasicPoolStats>(sharedManager, minMaxExpectedSparePoolResizer);
            break;
        case SIMPLE:
            break;
        }
        resizer.setResizable(pool);
        currentPool = pool;

        // 2 validator
        if (IValidatorPoolManager.class.isInstance(poolManager)) {
            IValidatorPoolManager<E> validatorCluePoolManager = (IValidatorPoolManager<E>) poolManager;
            validatorPool = new ValidatorPool<E>(currentPool, validatorCluePoolManager, false);
            currentPool = validatorPool;
        }

        // 3.0 stats & resize pool
        statsResizerPool = new StatsResizerPool<E, BasicPoolStats>(currentPool,
                new BasicPoolStatsFillerPoolListener<E>(), new BasicPoolStats(), resizer);
        // 3.1 add a 3e listener
        // FIXME manage remove listener and activate / passivate listener => create new optional managers
        // statsResizerCluePool.addCluePoolListener(poolFactoryImpl);
        currentPool = statsResizerPool;

        // 4 keep busy pool
        keepBusyPool = new KeepBusyPool<E>(currentPool);
        currentPool = keepBusyPool;
        // 5 limited waiter pool
        limitedWaiterPool = new LimitedWaiterPool<E>(currentPool);
        currentPool = limitedWaiterPool;
        // end pool construction

        this.limitedWaiterPool = limitedWaiterPool;
        this.keepBusyPool = keepBusyPool;
        this.lastPool = currentPool;

        keepBusyPool.setShareMethod(ShareMethod.NEVER_SHARE);
        PoolConfiguration poolConfiguration = new PoolConfiguration();
        int maxPool = poolConfiguration.getMax();
        timeout = poolConfiguration.getTimeout();
        maxWaiter = poolConfiguration.getMaxWaiters();
        limitedWaiterPool.setMaxWaiter(maxWaiter);
        minMaxExpectedSparePoolResizer.setMaxPool(maxPool);
        statsResizerPool.prepareUpdate();
    }

    protected IWaitControl createWaitControl() {
        if (timeout < 0) {
            throw new IllegalStateException();
        }
        if (timeout == 0) {
            return new InfiniteWaitControl();
        }
        return new ExpireWaitControl(System.currentTimeMillis() + timeout);
    }

    private Object sleepMutex = new Object();

    public E get() throws PoolException {
        do {
            try {
                LOG.debug("START GET {1} try to get with timeout {0}", timeout);
                E e = lastPool.get(createWaitControl());
                LOG.debug("OK FOR GET {1} try to get with timeout {0}", timeout);
                return e;
            } catch (TimeoutPoolException e) {
                throw new PoolException("No more instances available", e);
            } catch (TooManyWaiterException e) {
                throw new PoolException("No more instances available", e);
            } catch (PoolFactoryBroken e) {
                Long deadLine = e.getExpireBrokenDeadline();
                if (deadLine == null) {
                    throw new PoolException("Factory broken", e);
                } else {
                    synchronized (sleepMutex) {
                        long timeToWait = deadLine - System.currentTimeMillis();
                        while (timeToWait > 0) {
                            try {
                                sleepMutex.wait(timeToWait);
                            } catch (InterruptedException e1) {
                                throw new PoolException("Interrupted exception", e);
                            }
                            timeToWait = deadLine - System.currentTimeMillis();
                        }
                    }
                }
            } catch (org.ow2.util.pool.impl.enhanced.api.PoolException e) {
                throw new PoolException("Unknow exception", e);
            } catch (InterruptedException e) {
                throw new PoolException("Interrupted exception", e);
            }
        } while (true);
    }

    public void discard(final E instance) throws PoolException {
        LOG.debug("discard {0} ", instance);
        try {
            lastPool.remove(instance);
            basicPool.waitAllRemoveListenerCompleted();
        } catch (org.ow2.util.pool.impl.enhanced.api.PoolException e) {
            throw new PoolException("Unknow exception", e);
        } catch (InterruptedException e) {
            throw new PoolException("Interrupted exception", e);
        }
    }

    public void release(final E instance) throws PoolException {
        LOG.debug("release {0} ", instance);
        try {
            lastPool.put(instance);
        } catch (org.ow2.util.pool.impl.enhanced.api.NotABusyPoolItemException e) {
            LOG.debug("try to release a removed instance, ignore this request");
        } catch (org.ow2.util.pool.impl.enhanced.api.PoolException e) {
            throw new PoolException("Unknow exception", e);
        }
    }

    /**
     * If you do not want to wait an action which resize the pool to create future needed pool item.
     * This method can be use after a failed stop() to reconstruct missing items.
     * Also after construction and start() ...
     */
    public void updatePoolSize() {
        statsResizerPool.update();
    }

    public void setPoolConfiguration(final IPoolConfiguration poolConfiguration) {
        int maxPool = poolConfiguration.getMax();
        timeout = poolConfiguration.getTimeout();
        maxWaiter = poolConfiguration.getMaxWaiters();
        limitedWaiterPool.setMaxWaiter(maxWaiter);

        minMaxExpectedSparePoolResizer.setExpectedSparePool(poolConfiguration.getSpare());
        minMaxExpectedSparePoolResizer.setMinPool(poolConfiguration.getMin());
        minMaxExpectedSparePoolResizer.setMaxPool(maxPool);
        statsResizerPool.prepareUpdate();
        LOG.debug("{3} new conf Setted (timeout : {0}, maxwaiter: {1}, max: {2})", timeout, maxWaiter,
                maxPool, this);
    }

    /**
     * Allow to share instances (shouldn't be use in stateful case as one
     * stateful ID is linked to one client. No Stateful with same ID are
     * authorized).
     * @param allowSharedInstance true if it is allowed, else false.
     */
    public void setAllowSharedInstance(final boolean allowSharedInstance) {
        if (allowSharedInstance) {
            keepBusyPool.setShareMethod(ShareMethod.SHARE_AS_LAST_SOLUTION);
        } else {
            keepBusyPool.setShareMethod(ShareMethod.NEVER_SHARE);
        }
    }

    /**
     * Return true if many instances of the same object can be created (when
     * pool is not full).
     * @return true if this is allowed.
     */
    public boolean isAllowSharedInstance() {
        return keepBusyPool.getShareMethod() != ShareMethod.NEVER_SHARE;
    }

    public void start() throws PoolException {
        synchronized (mutex) {
            if (started) {
                throw new PoolException("Pool already started");
            }
            limitedWaiterPool.setMaxWaiter(maxWaiter);
            basicPool.unlockSizeToZero();
            started = true;
        }
    }

    public void forceStop() throws InterruptedException {
        createCB.forceStop(null);
        removeCB.forceStop(null);
    }


    public void stop() throws PoolException {
        synchronized (mutex) {
            if (!started) {
                throw new PoolException("Pool already stopped");
            }
            // all following update call will not create new item
            basicPool.lockSizeToZero();
            // new call to get will result in a TooManyWaiterException.
            limitedWaiterPool.setMaxWaiter(0);
            waitAuthorization.forbidWait();
            // interrupt current waiters
            lastPool.signalAllWaiters();
            // new get can wait however while maxWaiter is set to zero this
            // is still impossible
            waitAuthorization = new WaitAuthorization();
            try {
                // now all threads in pool are throwing exception or getting a
                // new item, so we just wait these work to terminate
                limitedWaiterPool.waitEmptyWaiterInPool();
            } catch (InterruptedException e) {
                basicPool.unlockSizeToZero();
                limitedWaiterPool.setMaxWaiter(maxWaiter);
                throw new PoolException("Interrupted exception", e);
            }
            // remove busy instances
            keepBusyPool.removeAll();
            try {
                // remove not busy instances, call setExpectedSize(0)
                basicPool.clearPool();
                // now all pool items must be removed from pool
                basicPool.waitAllRemoveListenerCompleted();
            } catch (InterruptedException e) {
                // undo operation
                basicPool.unlockSizeToZero();
                limitedWaiterPool.setMaxWaiter(maxWaiter);
                throw new PoolException("Interrupted exception", e);
            }
            started = false;
        }
    }


    public IPoolState getState() {
        BasicPoolState state = new BasicPoolState();
        basicPool.fillState(state);
        return state;
    }

    /**
     * @return pool configuration.
     */
    public IPoolConfiguration getPoolConfiguration() {
        PoolConfiguration poolConfiguration = new PoolConfiguration();
        poolConfiguration.setMin(minMaxExpectedSparePoolResizer.getMinPool());
        poolConfiguration.setMax(minMaxExpectedSparePoolResizer.getMaxPool());
        poolConfiguration.setSpare(minMaxExpectedSparePoolResizer.getExpectedSparePool());
        poolConfiguration.setMaxWaiters(limitedWaiterPool.getMaxWaiter());
        return poolConfiguration;
    }

}
